@@ -7,29 +7,23 @@ import hashlib
 import json
 from base64 import b64decode
 from collections.abc import AsyncIterator, Callable, Coroutine
-from contextvars import ContextVar
 from pathlib import Path
 from typing import Any, Optional
 
-import aiofiles
 import attr
 import httpx
 import typer
+from rich.progress import TaskID
 from yarl import URL
 
 from fourget import __version__, log
-from fourget.console import PROGRESS, PROGRESS_TASK
-from fourget.queue import Item, Queue
-
-client_var: ContextVar[httpx.AsyncClient] = ContextVar("client")
-result_var: ContextVar[Results] = ContextVar("results")
-
-
-@attr.s(auto_attribs=True, kw_only=True, order=False)
-class Results:
-    """Summary of processing."""
-
-    performed_new_download: bool = attr.ib(default=False)
+from fourget.console import PROGRESS
+from fourget.exception import (
+    FourgetException,
+    MalformedThreadURLException,
+    ThreadNotFoundException,
+)
+from fourget.queue import Item, Queue, StopReason
 
 
 @attr.s(frozen=True, auto_attribs=True, kw_only=True, order=False)
@@ -147,31 +141,40 @@ class Post:
 
 
 @attr.s(frozen=True, auto_attribs=True, kw_only=True, order=False)
-class ThreadURL:
-    """Data class for thread URLs."""
+class Thread:
+    """Data class for threads."""
 
     board: str
     thread_id: int
 
     @classmethod
-    def from_url(cls, url: str) -> ThreadURL:
+    def from_url(cls, url: str) -> Thread:
         """Create a new ThreadURL from a desktop URL."""
         parsed = URL(url)
         path = Path(parsed.path)
-        _, board, _, thread_id = path.parts
-        return ThreadURL(board=board, thread_id=int(thread_id))
+        try:
+            # value error may occur on unpack if not enough parts
+            _, board, _, thread_id = path.parts
+            # value error may occur if thread_id is not numeric
+            return Thread(board=board, thread_id=int(thread_id))
+        except ValueError as value_error:
+            raise MalformedThreadURLException(url) from value_error
 
     @property
     def api_endpoint_url(self) -> str:
         """Get the API endpoint URL."""
         return f"https://a.4cdn.org/{self.board}/thread/{self.thread_id}.json"
 
+    def to_url(self) -> str:
+        """Reconstruct the URL for this thread."""
+        return f"https://boards.4channel.org/{self.board}/thread/{self.thread_id}"
+
 
 async def md5_path(path: Path, max_chunk_size: int = 2 ** 20) -> bytes:
     """Return the md5 hash of a file from its path."""
-    async with aiofiles.open(path, "rb") as f:
+    with path.open("rb") as f:
         md5_hash = hashlib.md5()
-        while data := await f.read(max_chunk_size):
+        while data := f.read(max_chunk_size):
             md5_hash.update(data)
         return md5_hash.digest()
 
@@ -181,9 +184,9 @@ async def download_file(
 ) -> AsyncIterator[int]:
     """Download url to path with client, yielding chunk lengths as we go."""
     async with client.stream("GET", url) as response:
-        async with aiofiles.open(path, "wb") as f:
+        with path.open("wb") as f:
             async for chunk in response.aiter_bytes(max_chunk_size):
-                await f.write(chunk)
+                f.write(chunk)
                 yield len(chunk)
 
 
@@ -192,8 +195,11 @@ async def download_file(
 BLOCKED_FILE_NAME_CHARS = {chr(i) for i in range(32)} | set(R'<>:"/\|?*')
 
 
-def file_name_santize(name: str) -> str:
-    """Return a file name that should be suitable as a file name on most OSes."""
+def file_name_sanitize(name: str) -> str:
+    """
+    Return a file name that should be suitable on most OSes. Specifically, certain known
+    bad characters will be filtered out.
+    """
     return "".join(c for c in name if c not in BLOCKED_FILE_NAME_CHARS)
 
 
@@ -203,14 +209,18 @@ class MediaDownloadItem(Item):
 
     post_file: Post.File
     thread_dir: Path
+    progress_task: TaskID
+    client: httpx.AsyncClient
 
     async def process(
         self, enqueue: Callable[[Item], Coroutine[Any, Any, None]]
     ) -> None:
         """Download media files to local disk from URLs."""
-        local_file_name = file_name_santize(
-            f"{self.post_file.timestamp} - {self.post_file.poster_stem}"
-            f"{self.post_file.extension}"
+        local_file_name = (
+            file_name_sanitize(
+                f"{self.post_file.timestamp} - {self.post_file.poster_stem}"
+            )
+            + self.post_file.extension
         )
 
         output_path = self.thread_dir / local_file_name
@@ -220,40 +230,42 @@ class MediaDownloadItem(Item):
                 f"{self.post_file.url} already exists at "
                 f"{output_path.absolute().as_uri()}"
             )
-            PROGRESS.advance(PROGRESS_TASK, self.post_file.size)
+            PROGRESS.advance(self.progress_task, self.post_file.size)
         else:
-            client = client_var.get()
             async for chunk_size in download_file(
-                client=client, url=self.post_file.url, path=output_path
+                client=self.client, url=self.post_file.url, path=output_path
             ):
-                PROGRESS.advance(PROGRESS_TASK, chunk_size)
+                PROGRESS.advance(self.progress_task, chunk_size)
             log.info(f"{self.post_file.url} -> {output_path.absolute().as_uri()}")
-            result_var.get().performed_new_download = True
 
 
 @attr.s(frozen=True, auto_attribs=True, kw_only=True, order=False)
 class ThreadReadItem(Item):
     """Item that reads the thread."""
 
-    thread_url: ThreadURL
+    thread: Thread
     root_output_dir: Path
+    progress_task: TaskID
+    client: httpx.AsyncClient
 
     async def process(
         self, enqueue: Callable[[Item], Coroutine[Any, Any, None]]
     ) -> None:
         """Read the thread json and enqueue media downloads for posts with files."""
-        client = client_var.get()
-        response = await client.get(self.thread_url.api_endpoint_url)
+        response = await self.client.get(self.thread.api_endpoint_url)
+
+        if response.status_code == 404:
+            raise ThreadNotFoundException(self.thread.to_url())
 
         json_body = response.json()
 
         posts = [
-            Post.from_json(board=self.thread_url.board, json_resp=post)
+            Post.from_json(board=self.thread.board, json_resp=post)
             for post in json_body["posts"]
         ]
 
         total_file_size = sum(post.file.size for post in posts if post.file)
-        PROGRESS.update(PROGRESS_TASK, total=total_file_size)
+        PROGRESS.update(self.progress_task, total=total_file_size)
 
         orig_post = posts[0]
         assert orig_post
@@ -263,8 +275,8 @@ class ThreadReadItem(Item):
         else:
             trailer = f" - {orig_post.description}"
 
-        output_dir_name = file_name_santize(
-            f"4chan - {self.thread_url.board} - {orig_post.post_id}{trailer}"
+        output_dir_name = file_name_sanitize(
+            f"4chan - {self.thread.board} - {orig_post.post_id}{trailer}"
         )
 
         thread_dir = self.root_output_dir / output_dir_name
@@ -278,7 +290,14 @@ class ThreadReadItem(Item):
         for post in posts:
             if post.file is None:
                 continue
-            await enqueue(MediaDownloadItem(post_file=post.file, thread_dir=thread_dir))
+            await enqueue(
+                MediaDownloadItem(
+                    post_file=post.file,
+                    thread_dir=thread_dir,
+                    progress_task=self.progress_task,
+                    client=self.client,
+                )
+            )
 
 
 @attr.s(frozen=True, auto_attribs=True, kw_only=True, order=False)
@@ -294,46 +313,44 @@ class JSONSaveItem(Item):
         """Save the thread's json to 'thread.json' in the thread directory."""
         json_path = self.thread_path / "thread.json"
 
-        async with aiofiles.open(json_path, "w") as f:
-            await f.write(json.dumps(self.json_object, indent=2))
+        with json_path.open("w") as f:
+            f.write(json.dumps(self.json_object, indent=2))
 
         log.info(f"Saved thread json to {json_path.absolute().as_uri()}")
 
 
 async def start_queue(
-    thread_url: ThreadURL,
+    thread: Thread,
     root_output_dir: Path,
     queue_maxsize: int,
     worker_count: int,
-) -> bool:
+    progress_task: TaskID,
+) -> None:
     """Create a DownloadItem queue and start producers and consumers for it."""
-    log.info(f"Downloading files from {thread_url.api_endpoint_url}")
-
-    results = Results()
-    result_var.set(results)
+    log.info(f"Downloading files from {thread.api_endpoint_url}")
 
     queue: Queue = Queue.create(
         queue_maxsize=queue_maxsize,
         stop_reason_callback=lambda sr: log.info(str(sr)),
+        joined_stop_reason=StopReason(
+            reason=f"All media files of {thread.to_url()} have been gotten"
+        ),
     )
 
     async with httpx.AsyncClient() as client:
-        client_var.set(client)
-
         await queue.complete(
             initial_items=[
                 ThreadReadItem(
-                    thread_url=thread_url,
+                    thread=thread,
                     root_output_dir=root_output_dir,
+                    progress_task=progress_task,
+                    client=client,
                 ),
             ],
             worker_count=worker_count,
         )
 
-    PROGRESS.stop_task(PROGRESS_TASK)
-    PROGRESS.stop()
-
-    return results.performed_new_download
+    PROGRESS.stop_task(progress_task)
 
 
 app = typer.Typer()
@@ -391,17 +408,29 @@ def main(
     such as "https://boards.4channel.org/g/thread/76759434". The "4chan.org" domain may
     also be used.
     """
-    new_download = asyncio.run(
-        start_queue(
-            thread_url=ThreadURL.from_url(url),
-            root_output_dir=output_dir,
-            queue_maxsize=queue_maxsize,
-            worker_count=worker_count,
-        ),
-        debug=asyncio_debug,
-    )
-
-    raise typer.Exit(0 if new_download else 1)
+    # do initialization of the progress bar out here, because we need to reference it
+    # to clean it up when we're finished or an exception is raised
+    PROGRESS.start()
+    progress_task = PROGRESS.add_task(description="Downloading")
+    try:
+        asyncio.run(
+            start_queue(
+                thread=Thread.from_url(url),
+                root_output_dir=output_dir,
+                queue_maxsize=queue_maxsize,
+                worker_count=worker_count,
+                progress_task=progress_task,
+            ),
+            debug=asyncio_debug,
+        )
+    except FourgetException as fourget_exception:
+        log.error(fourget_exception.msg)
+        PROGRESS.update(progress_task, visible=False)
+        raise typer.Exit(1)
+    else:
+        raise typer.Exit(0)
+    finally:
+        PROGRESS.stop()
 
 
 if __name__ == "__main__":
