@@ -20,5 +20,3 @@ PROGRESS = Progress(
     TimeRemainingColumn(),
     console=Console(stderr=True),
 )
-PROGRESS.start()
-PROGRESS_TASK = PROGRESS.add_task(description="Downloading")
